==> g3dviewer-0.0.3.tar.gz: g3dviewer-0.0.3/src/object.c <==
#include <stdlib.h>
#include <math.h>

#include <glib.h>
#include <GL/gl.h>

#include <g3dviewer_plugin.h>

double object_radius(G3DObject *object)
{
  int i;
  double max_radius = 0.0;

  g_return_val_if_fail(object != NULL, 0.0);

  for(i=0; i<object->vertex_count; i++)
  {
    GLfloat *v = &object->vertices[i*3];
    double r = v[0]*v[0] + v[1]*v[1] + v[2]*v[2];
    if(r > max_radius) max_radius = r;
  }
#if DEBUG > 2
  g_printerr("cnt: %d, max_radius: %f, sqrt(): %f\n", object->vertex_count,
             max_radius, sqrt(max_radius));
#endif
  return sqrt(max_radius);
}

void object_scale(G3DObject *object, GLfloat scale)
{
  int i;

  for(i=0; i<object->vertex_count; i++)
  {
    object->vertices[i*3+0] *= scale;
    object->vertices[i*3+1] *= scale;
    object->vertices[i*3+2] *= scale;
  }
}

void object_clearmodel(G3DModel *model)
{
  GSList *list;

  /* lights */
  list = model->lights;
  while(list != NULL)
  {
    GSList *next;
    if(list->data != NULL) g_free(list->data);
    next = list->next;
    g_slist_free_1(list);
    list = next;
  }
  model->lights = NULL;

  /* objects */
  list = model->objects;
  while(list != NULL)
  {
    GSList *next, *slist;
    G3DObject *object = (G3DObject*)list->data;
    g_free(object->name);
    /* materials */
    slist = object->materials;
    while(slist != NULL)
    {
      GSList *snext;
      g_free(slist->data);
      snext = slist->next;
      g_slist_free_1(slist);
      slist = snext;
    }
    /* faces */
    slist = object->faces;
    while(slist != NULL)
    {
      GSList *snext;
      G3DFace *face = (G3DFace*)slist->data;
      if(face->index_vertex != NULL) g_free(face->index_vertex);
      if(face->texture_coords != NULL) g_free(face->texture_coords);
      g_free(face);
      snext = slist->next;
      g_slist_free_1(slist);
      slist = snext;
    }
    /* vertices */
    if(object->vertices != NULL) g_free(object->vertices);

    if(object->_normals != NULL) g_free(object->_normals);
    if(object->_indices != NULL) g_free(object->_indices);
    if(object->_materials != NULL) g_free(object->_materials);
    if(object->_flags != NULL) g_free(object->_flags);

    g_free(object);
    next = list->next;
    g_slist_free_1(list);
    list = next;
  }
  model->objects = NULL;

  list = model->materials;
  while(list != NULL)
  {
    GSList *next;
    if(list->data != NULL) g_free(list->data);
    next = list->next;
    g_slist_free_1(list);
    list = next;
  }
  model->materials = NULL;

  if(model->filename != NULL) g_free(model->filename);
  model->filename = NULL;
}

int object_centermodel(G3DModel *model)
{
  double min_x=10.0e99,min_y=10.0e99,min_z=10.0e99, 
         max_x=-9.9e99,max_y=-9.9e99,max_z=-9.9e99;
  double off_x, off_y, off_z;
  GSList *olist = model->objects;
  while(olist != NULL)
  {
    G3DObject *object = (G3DObject*)olist->data;
    int i;
    for(i=0; i<object->vertex_count; i++)
    {
      if(object->vertices[i*3+0] < min_x) min_x = object->vertices[i*3+0];
      if(object->vertices[i*3+1] < min_y) min_y = object->vertices[i*3+1];
      if(object->vertices[i*3+2] < min_z) min_z = object->vertices[i*3+2];
      if(object->vertices[i*3+0] > max_x) max_x = object->vertices[i*3+0];
      if(object->vertices[i*3+1] > max_y) max_y = object->vertices[i*3+1];
      if(object->vertices[i*3+2] > max_z) max_z = object->vertices[i*3+2];
    }
    olist = olist->next;
  }
  off_x = max_x - ((max_x - min_x) / 2.0);
  off_y = max_y - ((max_y - min_y) / 2.0);
  off_z = max_z - ((max_z - min_z) / 2.0);
#if DEBUG > 1
  g_printerr("x: %f - %f (%-f)\n", min_x, max_x, off_x);
  g_printerr("y: %f - %f (%-f)\n", min_y, max_y, off_y);
  g_printerr("z: %f - %f (%-f)\n", min_z, max_z, off_z);
#endif
  olist = model->objects;
  while(olist != NULL)
  {
    G3DObject *object = (G3DObject*)olist->data;
    int i;
    for(i=0; i<object->vertex_count; i++)
    {
      object->vertices[i*3+0] -= off_x;
      object->vertices[i*3+1] -= off_y;
      object->vertices[i*3+2] -= off_z;
    }
    olist = olist->next;
  }
  return EXIT_SUCCESS;
}

int object_optimize(G3DObject *object)
{
  unsigned int num_faces = 0, index = 0;
  GSList *flist = object->faces;
  int i, j;
  GLfloat nx, ny, nz;

  /* get the number of triangles in object */
  while(flist != NULL)
  {
    G3DFace *face = (G3DFace*)flist->data;
    if(face->num_vertices < 3)
    {
      flist = flist->next;
      continue;
    }
    num_faces += face->num_vertices - 2;
    flist = flist->next;
  }
  object->_num_faces = num_faces;
#if DEBUG > 1
  g_printerr("object_optimize: %d faces, %d triangles\n", 
             g_slist_length(object->faces), num_faces);
#endif
  
  /* allocate space for triangle indices */
  object->_normals = g_malloc(num_faces * 9 * sizeof(GLfloat));
  object->_materials = g_malloc0(num_faces * sizeof(G3DMaterial*));
  object->_flags = g_malloc0(num_faces * sizeof(unsigned int));
  object->_indices = g_malloc(num_faces * 3 * sizeof(unsigned int));

  /* copy face data to new arrays */
  flist = object->faces;
  while(flist != NULL)
  {
    G3DFace *face = (G3DFace*)flist->data;

    if(face->num_vertices < 3)
    {
      flist = flist->next;
      continue;
    }
    /* calculate default normal */
    if(!(face->flags & G3D_FLAG_FAC_NORMALS))
    {
      GLfloat ax, ay, az, bx, by, bz;

      ax = object->vertices[face->index_vertex[1]*3+0] - 
           object->vertices[face->index_vertex[0]*3+0];
      ay = object->vertices[face->index_vertex[1]*3+1] -
           object->vertices[face->index_vertex[0]*3+1];
      az = object->vertices[face->index_vertex[1]*3+2] -
           object->vertices[face->index_vertex[0]*3+2];
      bx = object->vertices[face->index_vertex[2]*3+0] -    
           object->vertices[face->index_vertex[0]*3+0];
      by = object->vertices[face->index_vertex[2]*3+1] -
           object->vertices[face->index_vertex[0]*3+1];
      bz = object->vertices[face->index_vertex[2]*3+2] -
           object->vertices[face->index_vertex[0]*3+2];

      g3d_vector_normal(ax, ay, az, bx, by, bz, &nx, &ny, &nz);

      if(g3d_vector_unify(&nx, &ny, &nz) != EXIT_SUCCESS)
      {
        flist = flist->next;
        object->_num_faces -= face->num_vertices - 2;
        continue;
      }
    }

    for(i=0; i<(face->num_vertices - 2); i++)
    {
      object->_materials[index] = face->material;
      object->_flags[index] = face->flags;

      for(j=0; j<3; j++)
      {
        if(j==0) object->_indices[index*3] = face->index_vertex[0];
        else object->_indices[index*3+j] = face->index_vertex[i+j];

        if(face->flags & G3D_FLAG_FAC_NORMALS)
        {
          object->_normals[(index*3+j)*3+0] = face->normals[(i+j)*3+0];
          object->_normals[(index*3+j)*3+1] = face->normals[(i+j)*3+1];
          object->_normals[(index*3+j)*3+2] = face->normals[(i+j)*3+2];
        }
        else
        {
          object->_normals[(index*3+j)*3+0] = nx;
          object->_normals[(index*3+j)*3+1] = ny;
          object->_normals[(index*3+j)*3+2] = nz;
        }
      }
      index++;
    }
    flist = flist->next;
  }

#if DEBUG > 1
  g_printerr("object_optimize: index %d, num_faces %d\n", 
             index, object->_num_faces);
#endif
  return EXIT_SUCCESS;
}

==> g3dviewer-0.1.1.tar.gz: g3dviewer-0.1.1/src/object.c <==
#include <stdlib.h>
#include <math.h>

#include <glib.h>
#include <GL/gl.h>

#include <g3dviewer_plugin.h>

double object_radius(G3DObject *object)
{
  int i;
  double max_radius = 0.0;

  g_return_val_if_fail(object != NULL, 0.0);

  for(i=0; i<object->vertex_count; i++)
  {
    GLfloat *v = &object->vertices[i*3];
    double r = v[0]*v[0] + v[1]*v[1] + v[2]*v[2];
    if(r > max_radius) max_radius = r;
  }
#if DEBUG > 2
  g_printerr("cnt: %d, max_radius: %f, sqrt(): %f\n", object->vertex_count,
             max_radius, sqrt(max_radius));
#endif
  return sqrt(max_radius);
}

void object_scale(G3DObject *object, GLfloat scale)
{
  int i;

  for(i=0; i<object->vertex_count; i++)
  {
    object->vertices[i*3+0] *= scale;
    object->vertices[i*3+1] *= scale;
    object->vertices[i*3+2] *= scale;
  }
}

G3DImage *free_G3DImage(G3DImage *image)
{
  if(image->pixeldata) g_free(image->pixeldata);
  if(image->name) g_free(image->name);
  g_free(image);
  return NULL;
}


G3DObject *free_G3DObject(G3DObject *object)
{
  GSList *slist;

  if(object->name != NULL) g_free(object->name);

  /* materials */
  slist = object->materials;
  while(slist != NULL)
  {
    GSList *snext;
    if(slist->data != NULL)
    {
      G3DMaterial *mat = (G3DMaterial*)slist->data;
      if(mat->texture != NULL) mat->texture = free_G3DImage(mat->texture);
      g_free(mat);
    }
    snext = slist->next;
    g_slist_free_1(slist);
    slist = snext;
  }
  /* faces */
  slist = object->faces;
  while(slist != NULL)
  {
    GSList *snext;
    G3DFace *face = (G3DFace*)slist->data;
    if(face->index_vertex != NULL) g_free(face->index_vertex);
    g_free(face);
    snext = slist->next;
    g_slist_free_1(slist);
    slist = snext;
  }
  /* vertices */
  if(object->vertices != NULL) g_free(object->vertices);
  if(object->tex_vertices != NULL) g_free(object->tex_vertices);
  if(object->_normals != NULL) g_free(object->_normals);
  if(object->_indices != NULL) g_free(object->_indices);
  if(object->_materials != NULL) g_free(object->_materials);
  if(object->_flags != NULL) g_free(object->_flags);

  g_free(object);
  return NULL;
}


void object_clearmodel(G3DModel *model)
{
  GSList *list;

  /* lights */
  list = model->lights;
  while(list != NULL)
  {
    GSList *next;
    if(list->data != NULL) g_free(list->data);
    next = list->next;
    g_slist_free_1(list);
    list = next;
  }
  model->lights = NULL;

  /* objects */
  list = model->objects;
  while(list != NULL)
  {
    GSList *next;
    G3DObject *object = (G3DObject*)list->data;

    free_G3DObject(object);

    next = list->next;
    g_slist_free_1(list);
    list = next;
  }
  model->objects = NULL;

  list = model->materials;
  while(list != NULL)
  {
    GSList *next;
    if(list->data != NULL) 
    {
      G3DMaterial *mat = (G3DMaterial*)list->data;
      if(mat->texture != NULL) 
      {
        mat->texture = free_G3DImage(mat->texture);
      }
      g_free(mat);
    }
    next = list->next;
    g_slist_free_1(list);
    list = next;
  }
  model->materials = NULL;

  if(model->filename != NULL) g_free(model->filename);
  model->filename = NULL;
}

int object_centermodel(G3DModel *model)
{
  double min_x=10.0e99,min_y=10.0e99,min_z=10.0e99, 
         max_x=-9.9e99,max_y=-9.9e99,max_z=-9.9e99;
  double off_x, off_y, off_z;
  GSList *olist = model->objects;
  while(olist != NULL)
  {
    G3DObject *object = (G3DObject*)olist->data;
    int i;
    for(i=0; i<object->vertex_count; i++)
    {
      if(object->vertices[i*3+0] < min_x) min_x = object->vertices[i*3+0];
      if(object->vertices[i*3+1] < min_y) min_y = object->vertices[i*3+1];
      if(object->vertices[i*3+2] < min_z) min_z = object->vertices[i*3+2];
      if(object->vertices[i*3+0] > max_x) max_x = object->vertices[i*3+0];
      if(object->vertices[i*3+1] > max_y) max_y = object->vertices[i*3+1];
      if(object->vertices[i*3+2] > max_z) max_z = object->vertices[i*3+2];
    }
    olist = olist->next;
  }
  off_x = max_x - ((max_x - min_x) / 2.0);
  off_y = max_y - ((max_y - min_y) / 2.0);
  off_z = max_z - ((max_z - min_z) / 2.0);
#if DEBUG > 1
  g_printerr("x: %f - %f (%-f)\n", min_x, max_x, off_x);
  g_printerr("y: %f - %f (%-f)\n", min_y, max_y, off_y);
  g_printerr("z: %f - %f (%-f)\n", min_z, max_z, off_z);
#endif
  olist = model->objects;
  while(olist != NULL)
  {
    G3DObject *object = (G3DObject*)olist->data;
    int i;
    for(i=0; i<object->vertex_count; i++)
    {
      object->vertices[i*3+0] -= off_x;
      object->vertices[i*3+1] -= off_y;
      object->vertices[i*3+2] -= off_z;
    }
    olist = olist->next;
  }
  return EXIT_SUCCESS;
}

int object_checkmodel(G3DModel *model)
{
	G3DFace *face;
	G3DObject *object;
	GSList *fitem, *oitem;
	guint32 i, no = 0, nf;

	oitem = model->objects;

	while(oitem)
	{
		object = (G3DObject *)oitem->data;

		nf = 0;
		fitem = object->faces;
		while(fitem)
		{
			face = (G3DFace *)fitem->data;

			if(face->material == NULL)
			{
				g_warning("object_check: face->material is NULL");
				return EXIT_FAILURE;
			}

			if(face->num_vertices < 3)
			{
				g_warning("object_check: face->num_vertices < 3");
				return EXIT_FAILURE;
			}

			for(i = 0; i < face->num_vertices; i ++)
			{
				if(face->index_vertex[i] >= object->vertex_count)
				{
					g_warning("object_check: [O: %d, F: %d] "
						"index_vertex[%d] >= vertex_count (%d >= %d)",
						no, nf,
						i, face->index_vertex[i], object->vertex_count);
					return EXIT_FAILURE;
				}
			}

			nf ++;
			fitem = fitem->next;
		} /* while(fitem) */

		no ++;
		oitem = oitem->next;
	} /* while(oitem) */

	return EXIT_SUCCESS;
}

int object_optimize(G3DObject *object)
{
  unsigned int num_faces = 0, index = 0;
  GSList *flist = object->faces;
  int i, j;
  GLfloat nx, ny, nz;

  /* get the number of triangles in object */
  while(flist != NULL)
  {
    G3DFace *face = (G3DFace*)flist->data;
	g_assert(face != NULL);
    if(face->num_vertices < 3)
    {
      flist = flist->next;
      continue;
    }
    num_faces += face->num_vertices - 2;
    flist = flist->next;
  }
  object->_num_faces = num_faces;
#if DEBUG > 1
  g_printerr("object_optimize: %d faces, %d triangles\n", 
             g_slist_length(object->faces), num_faces);
#endif
  
  /* allocate space for triangle indices */
  object->_normals = g_malloc(num_faces * 9 * sizeof(GLfloat));
  object->_materials = g_malloc0(num_faces * sizeof(G3DMaterial*));
  object->_flags = g_malloc0(num_faces * sizeof(unsigned int));
  object->_indices = g_malloc(num_faces * 3 * sizeof(unsigned int));

  /* copy face data to new arrays */
  flist = object->faces;
  while(flist != NULL)
  {
    G3DFace *face = (G3DFace*)flist->data;

    if(face->num_vertices < 3)
    {
      flist = flist->next;
      continue;
    }
    /* calculate default normal */
    if(!(face->flags & G3D_FLAG_FAC_NORMALS))
    {
      GLfloat ax, ay, az, bx, by, bz;

      ax = object->vertices[face->index_vertex[1]*3+0] - 
           object->vertices[face->index_vertex[0]*3+0];
      ay = object->vertices[face->index_vertex[1]*3+1] -
           object->vertices[face->index_vertex[0]*3+1];
      az = object->vertices[face->index_vertex[1]*3+2] -
           object->vertices[face->index_vertex[0]*3+2];
      bx = object->vertices[face->index_vertex[2]*3+0] -    
           object->vertices[face->index_vertex[0]*3+0];
      by = object->vertices[face->index_vertex[2]*3+1] -
           object->vertices[face->index_vertex[0]*3+1];
      bz = object->vertices[face->index_vertex[2]*3+2] -
           object->vertices[face->index_vertex[0]*3+2];

      g3d_vector_normal(ax, ay, az, bx, by, bz, &nx, &ny, &nz);

      if(g3d_vector_unify(&nx, &ny, &nz) != EXIT_SUCCESS)
      {
        flist = flist->next;
        object->_num_faces -= face->num_vertices - 2;
        continue;
      }
    }

    for(i=0; i<(face->num_vertices - 2); i++)
    {
      object->_materials[index] = face->material;
      object->_flags[index] = face->flags;

      for(j=0; j<3; j++)
      {
        if(j==0) object->_indices[index*3] = face->index_vertex[0];
        else object->_indices[index*3+j] = face->index_vertex[i+j];

        if(face->flags & G3D_FLAG_FAC_NORMALS)
        {
          object->_normals[(index*3+j)*3+0] = face->normals[(i+j)*3+0];
          object->_normals[(index*3+j)*3+1] = face->normals[(i+j)*3+1];
          object->_normals[(index*3+j)*3+2] = face->normals[(i+j)*3+2];
        }
        else
        {
          object->_normals[(index*3+j)*3+0] = nx;
          object->_normals[(index*3+j)*3+1] = ny;
          object->_normals[(index*3+j)*3+2] = nz;
        }
      }
      index++;
    }
    flist = flist->next;
  }

#if DEBUG > 1
  g_printerr("object_optimize: index %d, num_faces %d\n", 
             index, object->_num_faces);
#endif
  return EXIT_SUCCESS;
}
